﻿using DiscUtils;
using System;
using System.Collections.Generic;

namespace VHDXDump
{
    class Program
    {
        static string _vhdPath;
        static string _targetPath;

        static List<string> _tooLong = new List<string>();

        static void Main(string[] args)
        {
            if (args.Length < 3)
            {
                Console.WriteLine("ConsoleApp1.exe [vhd(x) path] [partition id] [target_path]");
                return;
            }

            _vhdPath = args[0];
            _partition = Int32.Parse(args[1]);
            _targetPath = args[2];

            Program program = new Program();
            program.Run();
        }

        private static int _partition = -1;

        protected static int Partition
        {
            get { return _partition; }
        }

        protected void Run()
        {
            VolumeManager volMgr = new VolumeManager();

            string uri = _vhdPath;
            volMgr.AddDisk(VirtualDisk.OpenDisk(uri, System.IO.FileAccess.Read));

            VolumeInfo volInfo = null;
            if (Partition >= 0)
            {
                volInfo = volMgr.GetPhysicalVolumes()[Partition];
            }
            else
            {
                volInfo = volMgr.GetLogicalVolumes()[0];
            }

            DiscUtils.FileSystemInfo fsInfo = FileSystemManager.DetectDefaultFileSystems(volInfo)[0];

            using (DiscFileSystem fs = fsInfo.Open(volInfo))
            {
                string targetFolder = _targetPath;
                if (System.IO.Directory.Exists(targetFolder) == false)
                {
                    System.IO.Directory.CreateDirectory(targetFolder);
                }

                CopyAllFilesTo(fs, "", targetFolder);
            }

            foreach (string txt in _tooLong)
            {
                Console.WriteLine("Excluded: " + txt);
            }
        }

        private void CopyAllFilesTo(DiscFileSystem fs, string srcPath, string dstPath)
        {
            string dstNewPath = System.IO.Path.Combine(dstPath, srcPath);
            if (System.IO.Directory.Exists(dstNewPath) == false)
            {
                try
                {
                    System.IO.Directory.CreateDirectory(dstNewPath);
                }
                catch (System.IO.PathTooLongException)
                {
                    _tooLong.Add(srcPath);
                    return;
                }
            }

            Console.WriteLine("get Sub Directories: " + srcPath);

            string[] dirs = fs.GetDirectories(srcPath);

            foreach (string dir in dirs)
            {
                string srcDirPath = System.IO.Path.Combine(srcPath, dir);
                CopyAllFilesTo(fs, dir, dstPath);
            }

            string[] files = fs.GetFiles(srcPath);

            try
            {
                foreach (string file in files)
                {
                    using (System.IO.Stream source = fs.OpenFile(file, System.IO.FileMode.Open, System.IO.FileAccess.Read))
                    {
                        string dstFilePath = System.IO.Path.Combine(dstPath, file);
                        using (System.IO.FileStream outFile = new System.IO.FileStream(dstFilePath, System.IO.FileMode.Create, System.IO.FileAccess.ReadWrite))
                        {
                            System.Console.WriteLine(file + " -> " + dstFilePath);
                            CopyStreams(source, outFile);
                        }
                    }
                }
            } catch (System.IO.PathTooLongException)
            {
                _tooLong.Add(srcPath);
            }
        }

        private void CopyAllFilesTo2(DiscFileSystem fs, string srcPath, string dstPath)
        {
            string dstNewPath = System.IO.Path.Combine(dstPath, srcPath);
            if (System.IO.Directory.Exists(dstNewPath) == false)
            {
                System.IO.Directory.CreateDirectory(dstNewPath);
            }

            Console.WriteLine("get Sub Directories: " + srcPath);

            string[] dirs = fs.GetDirectories(srcPath);

            foreach (string dir in dirs)
            {
                string srcDirPath = System.IO.Path.Combine(srcPath, dir);
                CopyAllFilesTo(fs, dir, dstPath);
            }

            string[] files = fs.GetFiles(srcPath);

            foreach (string file in files)
            {
                using (System.IO.Stream source = fs.OpenFile(file, System.IO.FileMode.Open, System.IO.FileAccess.Read))
                {
                    string dstFilePath = System.IO.Path.Combine(dstPath, file);
                    using (System.IO.FileStream outFile = new System.IO.FileStream(dstFilePath, System.IO.FileMode.Create, System.IO.FileAccess.ReadWrite))
                    {
                        CopyStreams(source, outFile);
                    }
                }
            }
        }

        private static void CopyStreams(System.IO.Stream inStream, System.IO.Stream outStream)
        {
            byte[] buffer = new byte[4096];
            int bytesRead = inStream.Read(buffer, 0, 4096);

            while (bytesRead != 0)
            {
                outStream.Write(buffer, 0, bytesRead);
                bytesRead = inStream.Read(buffer, 0, 4096);
            }
        }
    }
}
